"""Tasks for canvas"""

import hashlib
import logging
from functools import partial

import requests
from celery import shared_task
from django.contrib.auth import get_user_model
from lms.djangoapps.courseware.courses import get_course_by_id
from lms.djangoapps.grades.models import PersistentSubsectionGrade
from lms.djangoapps.instructor_task.api_helper import submit_task
from lms.djangoapps.instructor_task.tasks_base import BaseInstructorTask
from lms.djangoapps.instructor_task.tasks_helper.runner import run_main_task

from ol_openedx_canvas_integration import task_helpers
from ol_openedx_canvas_integration.api import get_subsection_user_grades
from ol_openedx_canvas_integration.client import CanvasClient, update_grade_payload_kv
from ol_openedx_canvas_integration.constants import (
    TASK_TYPE_PUSH_EDX_GRADES_TO_CANVAS,
    TASK_TYPE_SYNC_CANVAS_ENROLLMENTS,
)
from ol_openedx_canvas_integration.utils import get_canvas_course_id

TASK_LOG = logging.getLogger("edx.celery.task")
USER_MODEL = get_user_model()


def run_sync_canvas_enrollments(
    request, course_key, canvas_course_id, unenroll_current
):
    """
    Submit a task to start syncing canvas enrollments
    """
    task_type = TASK_TYPE_SYNC_CANVAS_ENROLLMENTS
    task_class = sync_canvas_enrollments_task
    task_input = {
        "course_key": course_key,
        "canvas_course_id": canvas_course_id,
        "unenroll_current": unenroll_current,
    }
    task_key = hashlib.md5(course_key.encode("utf8")).hexdigest()  # noqa: S324
    TASK_LOG.debug("Submitting task to sync canvas enrollments")
    return submit_task(request, task_type, task_class, course_key, task_input, task_key)


@shared_task(base=BaseInstructorTask)
def sync_canvas_enrollments_task(entry_id, xmodule_instance_args):
    """
    Fetch enrollments from canvas and update
    """
    action_name = "sync_canvas_enrollments"
    TASK_LOG.info("Running task to sync Canvas enrollments")
    task_fn = partial(task_helpers.sync_canvas_enrollments, xmodule_instance_args)
    return run_main_task(entry_id, task_fn, action_name)


def run_push_edx_grades_to_canvas(request, course_id):
    """
    Submit a task to start pushing edX grades to Canvas
    """
    task_type = TASK_TYPE_PUSH_EDX_GRADES_TO_CANVAS
    task_class = push_edx_grades_to_canvas_task
    task_input = {
        # course_key is already passed into the task, but we need to put it in task_input as well  # noqa: E501
        # so the instructor task status can be properly calculated instead of being marked incomplete  # noqa: E501
        "course_key": str(course_id)
    }
    task_key = hashlib.md5(course_id.encode("utf8")).hexdigest()  # noqa: S324
    TASK_LOG.debug("Submitting task to push edX grades to Canvas")
    return submit_task(request, task_type, task_class, course_id, task_input, task_key)


@shared_task(base=BaseInstructorTask)
def push_edx_grades_to_canvas_task(entry_id, xmodule_instance_args):
    """
    Push edX grades to Canvas
    """
    action_name = "push_edx_grades_to_canvas"
    TASK_LOG.info("Running task to push edX grades to Canvas")
    task_fn = partial(task_helpers.push_edx_grades_to_canvas, xmodule_instance_args)
    return run_main_task(entry_id, task_fn, action_name)


@shared_task
def sync_user_grade_with_canvas(grade_id):
    """
    Call the Canvas API and update the user's grade.
    """
    grade_instance = PersistentSubsectionGrade.objects.get(id=grade_id)
    course = get_course_by_id(grade_instance.course_id)
    canvas_course_id = get_canvas_course_id(course)
    if not canvas_course_id:
        TASK_LOG.debug("Canvas course ID not found. Skipping grade sync.")
        return

    TASK_LOG.info(
        "Course %s linked to Canvas Course %s. Attempting to sync grade.",
        grade_instance.course_id,
        canvas_course_id,
    )

    client = CanvasClient(canvas_course_id=canvas_course_id)
    existing_assignments_map = client.get_canvas_assignments()

    if str(grade_instance.full_usage_key) not in existing_assignments_map:
        TASK_LOG.warning(
            "The assignment %s is not synced with Canvas. Skipping grade sync.",
            grade_instance.usage_key,
        )
        return

    canvas_assignment_id = existing_assignments_map[str(grade_instance.usage_key)]["id"]
    openedx_user = USER_MODEL.objects.get(id=grade_instance.user_id)
    canvas_user_id = client.get_student_id_by_email(openedx_user.email)
    if not canvas_user_id:
        TASK_LOG.warning("The user %s is not enrolled in Canvas.", openedx_user)
        return

    # NOTE, theoretically we could simply use grade points from the ``grade_instance``.
    # However, the grade calculation seems a bit more complicated with overrides.
    # So, this uses the CourseGrade Python API.
    grade_dict = get_subsection_user_grades(
        course, grade_instance.full_usage_key, openedx_user
    )
    try:
        grade = grade_dict[grade_instance.full_usage_key][openedx_user]
        payload = dict([update_grade_payload_kv(canvas_user_id, grade.percent_graded)])
    except (KeyError, AttributeError):
        TASK_LOG.error(
            "Couldn't get grade for subsection <%s> with user <%s>",
            grade_instance.full_usage_key,
            grade_instance.user_id,
        )
        return

    res = client.update_assignment_grades(canvas_assignment_id, payload)

    if res.status_code != requests.codes.ok:
        TASK_LOG.error(
            "Canvas Integration: Grade Synced failed. HTTP Status: %s", res.status_code
        )
        return

    TASK_LOG.info("Canvas Integration: Grade synced successfully.")



"""Views for canvas integration"""

import logging

from common.djangoapps.student.models import CourseEnrollment, CourseEnrollmentAllowed
from common.djangoapps.util.json_request import JsonResponse
from django.contrib.auth.models import User
from django.db import transaction
from django.utils.translation import gettext as _
from django.views.decorators.cache import cache_control
from django.views.decorators.csrf import ensure_csrf_cookie
from django.views.decorators.http import require_POST
from lms.djangoapps.courseware.courses import get_course_by_id
from lms.djangoapps.instructor import permissions
from lms.djangoapps.instructor.views.api import require_course_permission
from lms.djangoapps.instructor_task.api_helper import AlreadyRunningError
from opaque_keys.edx.locator import CourseLocator

from ol_openedx_canvas_integration import tasks
from ol_openedx_canvas_integration.client import CanvasClient
from ol_openedx_canvas_integration.constants import COURSE_KEY_ID_EMPTY
from ol_openedx_canvas_integration.utils import get_canvas_course_id

log = logging.getLogger(__name__)


def _get_edx_enrollment_data(email, course_key):
    """Helper function to look up some info regarding whether a user with a email address is enrolled in edx"""  # noqa: D401, E501
    user = User.objects.filter(email=email).first()
    allowed = CourseEnrollmentAllowed.objects.filter(
        email=email, course_id=course_key
    ).first()

    return {
        "exists_in_edx": bool(user),
        "enrolled_in_edx": bool(
            user and CourseEnrollment.is_enrolled(user, course_key)
        ),
        "allowed_in_edx": bool(allowed),
    }


@ensure_csrf_cookie
@cache_control(no_cache=True, no_store=True, must_revalidate=True)
@require_course_permission(permissions.OVERRIDE_GRADES)
def list_canvas_enrollments(request, course_id):  # noqa: ARG001
    """
    Fetch enrollees for a course in canvas and list them
    """
    if not course_id:
        raise Exception(COURSE_KEY_ID_EMPTY)  # noqa: TRY002

    course_key = CourseLocator.from_string(course_id)
    course = get_course_by_id(course_key)
    canvas_course_id = get_canvas_course_id(course)

    if not canvas_course_id:
        msg = f"No canvas_course_id set for course: {course_id}"
        raise Exception(msg)  # noqa: TRY002

    client = CanvasClient(canvas_course_id=canvas_course_id)
    # WARNING: this will block the web thread
    enrollment_dict = client.list_canvas_enrollments()

    results = [
        {"email": email, **_get_edx_enrollment_data(email, course_key)}
        for email in sorted(enrollment_dict.keys())
    ]
    return JsonResponse(results)


@transaction.non_atomic_requests
@require_POST
@ensure_csrf_cookie
@cache_control(no_cache=True, no_store=True, must_revalidate=True)
@require_course_permission(permissions.OVERRIDE_GRADES)
def add_canvas_enrollments(request, course_id):
    """
    Fetches enrollees for a course in canvas and enrolls those emails in the course in edX
    """  # noqa: D401, E501
    unenroll_current = request.POST.get("unenroll_current", "").lower() == "true"
    course_key = CourseLocator.from_string(course_id)
    course = get_course_by_id(course_key)
    canvas_course_id = get_canvas_course_id(course)
    if not canvas_course_id:
        msg = f"No canvas_course_id set for course {course_id}"
        raise Exception(msg)  # noqa: TRY002

    try:
        tasks.run_sync_canvas_enrollments(
            request=request,
            course_key=course_id,
            canvas_course_id=canvas_course_id,
            unenroll_current=unenroll_current,
        )
        log.info("Syncing canvas enrollments for course %s", course_id)
        success_status = _("Syncing canvas enrollments")
        return JsonResponse({"status": success_status})
    except AlreadyRunningError:
        already_running_status = _(
            "Syncing canvas enrollments. See Pending Tasks below to view the status."
        )
        return JsonResponse({"status": already_running_status})


@ensure_csrf_cookie
@cache_control(no_cache=True, no_store=True, must_revalidate=True)
@require_course_permission(permissions.OVERRIDE_GRADES)
def list_canvas_assignments(request, course_id):  # noqa: ARG001
    """List Canvas assignments"""
    if not course_id:
        raise Exception(COURSE_KEY_ID_EMPTY)  # noqa: TRY002

    course_key = CourseLocator.from_string(course_id)
    course = get_course_by_id(course_key)
    canvas_course_id = get_canvas_course_id(course)
    if not canvas_course_id:
        msg = f"No canvas_course_id set for course: {course_id}"
        raise Exception(msg)  # noqa: TRY002

    client = CanvasClient(canvas_course_id=canvas_course_id)
    return JsonResponse(client.list_canvas_assignments())


@ensure_csrf_cookie
@cache_control(no_cache=True, no_store=True, must_revalidate=True)
@require_course_permission(permissions.OVERRIDE_GRADES)
def list_canvas_grades(request, course_id):
    """List grades"""
    if not course_id:
        raise Exception(COURSE_KEY_ID_EMPTY)  # noqa: TRY002

    assignment_id = int(request.GET.get("assignment_id"))
    course_key = CourseLocator.from_string(course_id)
    course = get_course_by_id(course_key)
    canvas_course_id = get_canvas_course_id(course)
    if not canvas_course_id:
        msg = f"No canvas_course_id set for course {course_id}"
        raise Exception(msg)  # noqa: TRY002

    client = CanvasClient(canvas_course_id=canvas_course_id)
    return JsonResponse(client.list_canvas_grades(assignment_id=assignment_id))


@transaction.non_atomic_requests
@require_POST
@ensure_csrf_cookie
@cache_control(no_cache=True, no_store=True, must_revalidate=True)
@require_course_permission(permissions.OVERRIDE_GRADES)
def push_edx_grades(request, course_id):
    """Push user grades for all graded items in edX to Canvas"""
    if not course_id:
        raise Exception(COURSE_KEY_ID_EMPTY)  # noqa: TRY002

    course_key = CourseLocator.from_string(course_id)
    course = get_course_by_id(course_key)
    canvas_course_id = get_canvas_course_id(course)
    if not canvas_course_id:
        msg = f"No canvas_course_id set for course: {course_id}"
        raise Exception(msg)  # noqa: TRY002
    try:
        tasks.run_push_edx_grades_to_canvas(
            request=request,
            course_id=course_id,
        )
        log.info("Pushing edX grades to canvas for course %s", course_id)
        success_status = _("Pushing edX grades to canvas")
        return JsonResponse({"status": success_status})
    except AlreadyRunningError:
        already_running_status = _(
            "Pushing edX grades to canvas. See Pending Tasks below to view the status."
        )
        return JsonResponse({"status": already_running_status})



"""Utility functions for canvas integration"""

import logging
from collections import defaultdict

from common.djangoapps.student.models import CourseEnrollment, CourseEnrollmentAllowed
from django.contrib.auth.models import User
from lms.djangoapps.courseware.access import has_access
from lms.djangoapps.courseware.courses import get_course_by_id
from lms.djangoapps.grades.context import grading_context_for_course
from lms.djangoapps.grades.course_grade_factory import CourseGradeFactory
from opaque_keys.edx.locator import CourseLocator

from ol_openedx_canvas_integration.client import (
    CanvasClient,
    create_assignment_payload,
    update_grade_payload_kv,
)
from ol_openedx_canvas_integration.constants import COURSE_KEY_ID_EMPTY
from ol_openedx_canvas_integration.utils import get_canvas_course_id

log = logging.getLogger(__name__)


def first_or_none(iterable):
    """Returns the first item in the given iterable, or None if the iterable is empty"""  # noqa: D401
    return next((x for x in iterable), None)


def course_graded_items(course):
    grading_context = grading_context_for_course(course)
    for graded_item_type, graded_items in grading_context[
        "all_graded_subsections_by_type"
    ].items():
        for graded_item_index, graded_item in enumerate(graded_items, start=1):
            yield graded_item_type, graded_item, graded_item_index


def get_enrolled_non_staff_users(course):
    """
    Returns an iterable of non-staff enrolled users for a given course
    """  # noqa: D401
    return [
        user
        for user in CourseEnrollment.objects.users_enrolled_in(course.id)
        if not has_access(user, "staff", course)
    ]


def enroll_emails_in_course(emails, course_key):
    """
    Attempts to enroll all provided emails in a course. Emails without a corresponding
    user have a CourseEnrollmentAllowed object created for the course.
    """  # noqa: D401
    results = {}
    for email in emails:
        user = User.objects.filter(email=email).first()
        result = ""
        if not user:
            _, created = CourseEnrollmentAllowed.objects.get_or_create(
                email=email, course_id=course_key
            )
            if created:
                result = "User does not exist - created course enrollment permission"
            else:
                result = "User does not exist - enrollment is already allowed"
        elif not CourseEnrollment.is_enrolled(user, course_key):
            try:
                CourseEnrollment.enroll(user, course_key)
                result = "Enrolled user in the course"
            except Exception as ex:  # pylint: disable=broad-except  # noqa: BLE001
                result = f"Failed to enroll - {ex}"
        else:
            result = "User already enrolled"
        results[email] = result
    return results


def get_subsection_user_grades(course, subsection_usage_key=None, user=None):
    """
    Builds a dict of user grades grouped by block locator. Only returns grades if the assignment has been attempted
    by the given user.

    Args:
        course: The course object (of the type returned by courseware.courses.get_course_by_id)
        subsection_usage_key (BlockUsageLocator): subsection to filter the grades
        user: specific user to filter the grades

    Returns:
        dict: Block locators for graded items (assignments, exams, etc.) mapped to a dict of users
            and their grades for those assignments.
            Example: {
                <BlockUsageLocator for graded item>: {
                    <User object for student 1>: <grades.subsection_grade.CreateSubsectionGrade object>,
                    <User object for student 2>: <grades.subsection_grade.CreateSubsectionGrade object>,
                }
            }
    """  # noqa: D401, E501
    if user:
        enrolled_students = [user]
    else:
        enrolled_students = CourseEnrollment.objects.users_enrolled_in(course.id)

    subsection_grade_dict = defaultdict(dict)
    for student, course_grade, _error in CourseGradeFactory().iter(
        users=enrolled_students, course=course
    ):
        for subsection_dict in course_grade.graded_subsections_by_format().values():
            for subsection_block_locator, subsection_grade in subsection_dict.items():
                # Filter when subsection is specified
                if (
                    subsection_usage_key
                    and subsection_block_locator != subsection_usage_key
                ):
                    continue

                subsection_grade_dict[subsection_block_locator].update(
                    # Only include grades if the assignment/exam/etc. has been attempted
                    {student: subsection_grade}
                    if subsection_grade.graded_total.first_attempted
                    else {}
                )
    return subsection_grade_dict


def get_subsection_block_user_grades(course):
    """
    Builds a dict of user grades grouped by the subsection XBlock representing each graded item.
    Only returns grades if the assignment has been attempted by the given user.

    Args:
        course: The course object (of the type returned by courseware.courses.get_course_by_id)

    Returns:
        dict: Block objects representing graded items (assignments, exams, etc.) mapped to a dict of users
            and their grades for those assignments.
            Example: {
                <content.block_structure.block_structure.BlockData object for graded item>: {
                    <User object for student 1>: <grades.subsection_grade.CreateSubsectionGrade object>,
                    <User object for student 2>: <grades.subsection_grade.CreateSubsectionGrade object>,
                }
            }
    """  # noqa: D401, E501
    subsection_user_grades = get_subsection_user_grades(course)
    graded_subsection_blocks = [
        graded_item.get("subsection_block")
        for graded_item_type, graded_item, graded_item_index in course_graded_items(
            course
        )
    ]
    locator_block_dict = {
        block_locator: first_or_none(
            block
            for block in graded_subsection_blocks
            if block.location == block_locator
        )
        for block_locator in subsection_user_grades
    }
    return {
        block: subsection_user_grades[block_locator]
        for block_locator, block in locator_block_dict.items()
        if block is not None
    }


def sync_canvas_enrollments(course_key, canvas_course_id, unenroll_current):
    """
    Fetch enrollments from canvas and update

    Args:
        course_key (str): The edX course key
        canvas_course_id (int): The canvas course id
        unenroll_current (bool): If true, unenroll existing students if not staff
    """
    if not course_key:
        raise Exception(COURSE_KEY_ID_EMPTY)  # noqa: TRY002

    if not canvas_course_id:
        msg = f"No canvas_course_id set for course: {course_key}"
        raise Exception(msg)  # noqa: TRY002

    course_key = CourseLocator.from_string(course_key)
    course = get_course_by_id(course_key)

    client = CanvasClient(canvas_course_id)
    emails_to_enroll = client.list_canvas_enrollments()
    users_to_unenroll = []

    if unenroll_current:
        enrolled_user_dict = {
            user.email: user for user in get_enrolled_non_staff_users(course)
        }
        emails_to_enroll_set = set(emails_to_enroll)
        already_enrolled_email_set = set(enrolled_user_dict.keys())
        emails_to_enroll = emails_to_enroll_set - already_enrolled_email_set
        users_to_unenroll = [
            enrolled_user_dict[email]
            for email in (already_enrolled_email_set - emails_to_enroll)
        ]

    enrolled = enroll_emails_in_course(emails=emails_to_enroll, course_key=course_key)
    log.info("Enrolled users in course %s: %s", course_key, enrolled)

    if users_to_unenroll:
        for user_to_unenroll in users_to_unenroll:
            CourseEnrollment.unenroll(user_to_unenroll, course.id)
        log.info(
            "Unenrolled non-staff users in course %s: %s", course_key, users_to_unenroll
        )


def push_edx_grades_to_canvas(course):
    """
    Gathers all student grades for each assignment in the given course, creates equivalent assignment in Canvas
    if they don't exist already, and adds/updates the student grades for those assignments in Canvas.

    Args:
        course: The course object (of the type returned by courseware.courses.get_course_by_id)

    Returns:
        dict: A dictionary with some information about the success/failure of the updates
    """  # noqa: E501
    if not course:
        raise Exception(COURSE_KEY_ID_EMPTY)  # noqa: TRY002

    canvas_course_id = get_canvas_course_id(course)
    if not canvas_course_id:
        msg = f"No canvas_course_id set for course: {course.id}"
        raise Exception(msg)  # noqa: TRY002

    client = CanvasClient(canvas_course_id=canvas_course_id)
    existing_assignment_dict = client.get_canvas_assignments()
    subsection_block_user_grades = get_subsection_block_user_grades(course)

    # Populate missing assignments
    new_assignment_blocks = (
        subsection_block
        for subsection_block in subsection_block_user_grades
        if str(subsection_block.location) not in existing_assignment_dict
    )
    created_assignments = {
        subsection_block: client.create_canvas_assignment(
            create_assignment_payload(subsection_block)
        )
        for subsection_block in new_assignment_blocks
    }

    # Build request payloads for updating grades in each assignment
    enrolled_user_dict = client.list_canvas_enrollments()
    grade_update_payloads = {}
    for subsection_block, user_grade_dict in subsection_block_user_grades.items():
        grade_update_payloads[subsection_block] = dict(
            update_grade_payload_kv(
                enrolled_user_dict[student_user.email.lower()], grade.percent_graded
            )
            for student_user, grade in user_grade_dict.items()
            # Only add the grade if the user exists in Canvas
            if student_user.email.lower() in enrolled_user_dict
        )

    # Send requests to update grades in each relevant course
    assignment_grades_updated = {
        subsection_block: client.update_assignment_grades(
            canvas_assignment_id=existing_assignment_dict[
                str(subsection_block.location)
            ]["id"],
            payload=grade_request_payload,
        )
        for subsection_block, grade_request_payload in grade_update_payloads.items()
        if grade_request_payload
        and str(subsection_block.location) in existing_assignment_dict
    }

    return assignment_grades_updated, created_assignments



DEFAULT_ASSIGNMENT_POINTS = 1
TASK_TYPE_SYNC_CANVAS_ENROLLMENTS = "sync_canvas_enrollments"
TASK_TYPE_PUSH_EDX_GRADES_TO_CANVAS = "push_edx_grades_to_canvas"
CANVAS_TASK_TYPES = [
    TASK_TYPE_SYNC_CANVAS_ENROLLMENTS,
    TASK_TYPE_PUSH_EDX_GRADES_TO_CANVAS,
]
COURSE_KEY_ID_EMPTY = "Course Key/ID is empty."



"""
Helper functions for canvas integration tasks
"""

import datetime
from time import time

import pytz
from lms.djangoapps.courseware.courses import get_course_by_id
from lms.djangoapps.instructor_task.api import get_running_instructor_tasks
from lms.djangoapps.instructor_task.models import InstructorTask
from lms.djangoapps.instructor_task.tasks_helper.runner import TaskProgress

from ol_openedx_canvas_integration import api
from ol_openedx_canvas_integration.constants import CANVAS_TASK_TYPES


def sync_canvas_enrollments(
    _xmodule_instance_args,
    _entry_id,
    course_id,  # noqa: ARG001
    task_input,
    action_name,
):
    """Partial function to sync canvas enrollments"""
    start_time = time()
    num_reports = 1
    task_progress = TaskProgress(action_name, num_reports, start_time)
    api.sync_canvas_enrollments(
        course_key=task_input["course_key"],
        canvas_course_id=task_input["canvas_course_id"],
        unenroll_current=task_input["unenroll_current"],
    )
    # for simplicity, only one task update for now when everything is done
    return task_progress.update_task_state(extra_meta={"step": "Done"})


def push_edx_grades_to_canvas(
    _xmodule_instance_args,
    _entry_id,
    course_id,
    task_input,  # noqa: ARG001
    action_name,
):
    """Partial function to push edX grades to canvas"""
    start_time = time()
    num_reports = 1
    task_progress = TaskProgress(action_name, num_reports, start_time)
    course = get_course_by_id(course_id)
    grades_updated, assignments_created = api.push_edx_grades_to_canvas(course=course)
    results = {"grades": len(grades_updated), "assignments": len(assignments_created)}
    return task_progress.update_task_state(
        extra_meta={"step": "Done", "results": results}
    )


def get_filtered_instructor_tasks(course_id, user):
    """
    Return a filtered query of InstructorTasks based on the course, user, and desired
    task types
    """
    instructor_tasks = get_running_instructor_tasks(course_id)
    now = datetime.datetime.now(pytz.utc)
    filtered_tasks = InstructorTask.objects.filter(
        course_id=course_id,
        task_type__in=CANVAS_TASK_TYPES,
        updated__lte=now,
        updated__gte=now - datetime.timedelta(days=2),
        requester=user,
    ).order_by("-updated")

    return (instructor_tasks | filtered_tasks).distinct()[0:3]



"""Django Signal handlers."""

import logging

from ol_openedx_canvas_integration.tasks import sync_user_grade_with_canvas

log = logging.getLogger(__name__)


def update_grade_in_canvas(sender, instance, created, **kwargs):  # noqa: ARG001
    """
    Automatically update grades in Canvas when assignments are synced.

    This signal receiver is wired to the `post_save` signal from the
    lms.djangoapps.grades.models.PersistentSubsectionGrade model. It then
    updates the Canvas course if the subsection is already synced to a Canvas
    course as an assignment.
    """
    log.debug("Grade updated, triggering background task")
    sync_user_grade_with_canvas.delay(instance.id)



"""Common settings unique to the canvas integration plugin."""


def plugin_settings(settings):
    """Settings for the canvas integration plugin."""  # noqa: D401
    settings.CANVAS_ACCESS_TOKEN = None
    settings.CANVAS_BASE_URL = None



"""
Canvas Integration Application Configuration
"""

from django.apps import AppConfig
from edx_django_utils.plugins import (
    PluginContexts,
    PluginSettings,
    PluginSignals,
    PluginURLs,
)
from lms.djangoapps.instructor.constants import INSTRUCTOR_DASHBOARD_PLUGIN_VIEW_NAME
from openedx.core.constants import COURSE_ID_PATTERN
from openedx.core.djangoapps.plugins.constants import ProjectType, SettingsType


class CanvasIntegrationConfig(AppConfig):
    """
    Configuration class for Canvas integration app
    """

    name = "ol_openedx_canvas_integration"

    plugin_app = {
        PluginURLs.CONFIG: {
            ProjectType.LMS: {
                PluginURLs.NAMESPACE: "",
                PluginURLs.REGEX: f"courses/{COURSE_ID_PATTERN}/canvas/api/",
                PluginURLs.RELATIVE_PATH: "urls",
            }
        },
        PluginSettings.CONFIG: {
            ProjectType.LMS: {
                SettingsType.PRODUCTION: {
                    PluginSettings.RELATIVE_PATH: "settings.production"
                },
                SettingsType.COMMON: {PluginSettings.RELATIVE_PATH: "settings.common"},
            },
            ProjectType.CMS: {
                SettingsType.PRODUCTION: {
                    PluginSettings.RELATIVE_PATH: "settings.production"
                },
                SettingsType.COMMON: {PluginSettings.RELATIVE_PATH: "settings.common"},
            },
        },
        PluginContexts.CONFIG: {
            ProjectType.LMS: {
                INSTRUCTOR_DASHBOARD_PLUGIN_VIEW_NAME: "ol_openedx_canvas_integration.context_api.plugin_context"  # noqa: E501
            }
        },
        PluginSignals.CONFIG: {
            ProjectType.LMS: {
                PluginSignals.RELATIVE_PATH: "receivers",
                PluginSignals.RECEIVERS: [
                    {
                        PluginSignals.RECEIVER_FUNC_NAME: "update_grade_in_canvas",
                        PluginSignals.SIGNAL_PATH: "django.db.models.signals.post_save",
                        PluginSignals.SENDER_PATH: "lms.djangoapps.grades.models.PersistentSubsectionGrade",  # noqa: E501
                    }
                ],
            }
        },
    }

    def ready(self):
        """Perform initialization tasks required for the plugin."""
        from ol_openedx_canvas_integration import handlers  # noqa: F401, PLC0415



"""Utilities for Canvas plugin"""


def get_canvas_course_id(course=None):
    """Get the course Id from the course settings"""
    return course.other_course_settings.get("canvas_id") if course else None


def get_task_output_formatted_message(task_output):
    """Take the edX task output and format a message for table display on task result"""
    # this reports on actions for a course as a whole
    results = task_output.get("results", {})
    assignments_count = results.get("assignments", 0)
    grades_count = results.get("grades", 0)

    return (
        f"{grades_count} grades and {assignments_count} assignments updated or created"
    )



"""
The initialization of the context for the Canvas Integration Plugin
"""

import pkg_resources
from django.urls import reverse
from django.utils.translation import gettext as _
from web_fragments.fragment import Fragment

from ol_openedx_canvas_integration.utils import get_canvas_course_id


def get_resource_bytes(path):
    """
    Helper method to get the unicode contents of a resource in this repo.

    Args:
        path (str): The path of the resource

    Returns:
        unicode: The unicode contents of the resource at the given path
    """  # noqa: D401
    resource_contents = pkg_resources.resource_string(__name__, path)
    return resource_contents.decode("utf-8")


def plugin_context(context):
    """Provide context based data for Canvas Integration plugin (For Instructor Dashboard)"""  # noqa: E501

    course = context.get("course")

    # Don't add Canvas tab is the Instructor Dashboard if it doesn't have any associated
    # canvas_course_id set from Canvas Service
    if not get_canvas_course_id(course):
        return None

    fragment = Fragment()
    # Adding JS as bytes (Inspired by what we are doing with Rapid Response xBlock)
    fragment.add_javascript(get_resource_bytes("static/js/canvas_integration.js"))

    canvas_context = {
        "section_key": "canvas_integration",
        "section_display_name": _("Canvas"),
        "course": context["course"],
        "add_canvas_enrollments_url": reverse(
            "add_canvas_enrollments", kwargs={"course_id": course.id}
        ),
        "list_canvas_enrollments_url": reverse(
            "list_canvas_enrollments", kwargs={"course_id": course.id}
        ),
        "list_canvas_assignments_url": reverse(
            "list_canvas_assignments", kwargs={"course_id": course.id}
        ),
        "list_canvas_grades_url": reverse(
            "list_canvas_grades", kwargs={"course_id": course.id}
        ),
        "list_instructor_tasks_url": "{}?include_canvas=true".format(
            reverse("list_instructor_tasks", kwargs={"course_id": course.id})
        ),
        "push_edx_grades_url": reverse(
            "push_edx_grades", kwargs={"course_id": course.id}
        ),
        "fragment": fragment,
        "template_path_prefix": "/",
    }

    sections = context.get("sections", [])
    sections.append(canvas_context)
    context["sections"] = sections

    return context



"""Event Handlers for the openedx-events signals."""

from django.dispatch import receiver
from openedx_events.content_authoring.signals import XBLOCK_DELETED, XBLOCK_PUBLISHED

from ol_openedx_canvas_integration.cms_tasks import sync_course_assignments_with_canvas


@receiver(XBLOCK_PUBLISHED)
def handle_xblock_publised_event(signal, sender, xblock_info, metadata, **kwargs):  # noqa: ARG001
    """
    Update Canvas assignments if the courses are linked.
    """
    sync_course_assignments_with_canvas.delay(str(xblock_info.usage_key.course_key))


@receiver(XBLOCK_DELETED)
def handle_xblock_deleted_event(signal, sender, xblock_info, metadata, **kwargs):  # noqa: ARG001
    """
    Update Canvas assignments if the courses are linked.
    """
    if xblock_info.block_type in ["chapter", "sequential"]:
        # Sometimes there is a race condition where the Canvas API calls are made
        # before the Course Outline has updated, leading to the assignments not getting
        # deleted in Canvas. So, a countdown of 10 seconds set to avoid it.
        sync_course_assignments_with_canvas.apply_async(
            args=[str(xblock_info.usage_key.course_key)],
            countdown=10,
        )



import logging
from urllib.parse import parse_qs, urlencode, urljoin, urlparse

import pytz
import requests
from django.conf import settings
from django.core.cache import cache

from ol_openedx_canvas_integration.constants import DEFAULT_ASSIGNMENT_POINTS

log = logging.getLogger(__name__)


class CanvasClient:
    def __init__(self, canvas_course_id):
        self.session = self.get_canvas_session()
        self.canvas_course_id = canvas_course_id

    @staticmethod
    def get_canvas_session():
        """
        Create a request session with the access token
        """
        session = requests.Session()
        session.headers.update(
            {"Authorization": f"Bearer {settings.CANVAS_ACCESS_TOKEN}"}
        )
        return session

    @staticmethod
    def _add_per_page(url, per_page):
        """
        Add per_page query parameter to override default value of 10

        Args:
            url (str): The url to update
            per_page (int): The new per_page value

        Returns:
            str: The updated URL
        """
        pieces = urlparse(url)
        query = parse_qs(pieces.query)
        query["per_page"] = per_page
        query_string = urlencode(query, doseq=True)
        pieces = pieces._replace(query=query_string)
        return pieces.geturl()

    def _paginate(self, url, *args, **kwargs):
        """
        Iterate over the paginated results of a request
        """
        url = self._add_per_page(
            url, 100
        )  # increase per_page to 100 from default of 10

        items = []
        while url:
            resp = self.session.get(url, *args, **kwargs)
            resp.raise_for_status()
            items.extend(resp.json())
            links = requests.utils.parse_header_links(resp.headers["link"])
            url = None
            for link in links:
                if link["rel"] == "next":
                    url = link["url"]

        return items

    def list_canvas_enrollments(self):
        """
        Fetch canvas enrollments. This may take a while, so don't run in the request thread.

        Returns:
            dict: Email addresses mapped to canvas user ids for all enrolled users
        """  # noqa: E501
        url = urljoin(
            settings.CANVAS_BASE_URL,
            f"/api/v1/courses/{self.canvas_course_id}/enrollments",
        )
        enrollments = self._paginate(url)

        return {
            enrollment["user"]["login_id"].lower(): enrollment["user"]["id"]
            for enrollment in enrollments
        }

    def list_canvas_assignments(self):
        """
        List Canvas assignments

        Returns:
            list: A list of assignment dicts from Canvas
        """
        url = urljoin(
            settings.CANVAS_BASE_URL,
            f"/api/v1/courses/{self.canvas_course_id}/assignments",
        )
        return self._paginate(url)

    def get_student_id_by_email(self, email: str):
        """
        Get the canvas ID of the learner with the given email.

        Returns:
            int: Canvas ID of the student if enrolled in the course. None otherwise.
        """
        key = "canvas-id-" + email
        if student_id := cache.get(key):
            return student_id

        url = urljoin(
            settings.CANVAS_BASE_URL,
            f"/api/v1/courses/{self.canvas_course_id}/search_users",
        )
        search_results = self._paginate(
            url, params={"search_term": email, "enrollment_type[]": "student"}
        )
        student_id = next(
            (
                user["id"]
                for user in search_results
                if user.get("email", "").lower() == email.lower()
            ),
            None,
        )
        if student_id:
            cache.set(key, student_id)
        return student_id

    def get_canvas_assignments(self):
        """
        Get Canvas assignments organized by integration_id.

        Returns:
            dict: A dictionary mapping integration_id to assignment data containing
                  'id' and 'is_published' fields. Only includes assignments with
                  integration_id set.
        """
        assignments = self.list_canvas_assignments()
        assignments_dict = {
            assignment.get("integration_id"): {
                "id": assignment["id"],
                "is_published": assignment.get("published", False),
            }
            for assignment in assignments
            if assignment.get("integration_id") is not None
        }
        assignments_without_integration_id = sorted(
            assignment["id"]
            for assignment in assignments
            if assignment.get("integration_id") is None
        )
        if assignments_without_integration_id:
            log.warning(
                "These assignments are missing an integration_id: %s",
                ", ".join(
                    str(assignment_id)
                    for assignment_id in assignments_without_integration_id
                ),
            )
        return assignments_dict

    def list_canvas_grades(self, assignment_id):
        """
        List grades for a Canvas assignment

        Args:
            assignment_id (int): The canvas assignment id
        """
        url = urljoin(
            settings.CANVAS_BASE_URL,
            f"/api/v1/courses/{self.canvas_course_id}/assignments/{assignment_id}/submissions",
        )
        return self._paginate(url)

    def create_canvas_assignment(self, payload):
        """
        Create an assignment on Canvas

        Args:
            payload (dict):
        """
        return self.session.post(
            url=urljoin(
                settings.CANVAS_BASE_URL,
                f"/api/v1/courses/{self.canvas_course_id}/assignments",
            ),
            json=payload,
        )

    def update_canvas_assignment(self, assignment_id, payload):
        """
        Update an assignment with the given ID in Canvas.

        Args:
            assignment_id (int): ID of the Canvas assignment
            payload (dict): Assignment data to update in Canvas

        Example payload:
            {
              "assignment": {
                "name": "Midterm Exam",
                "integration_id": "block-v1:Org+CS101type@sequential+block@exam",
                "grading_type": "percent",
                "points_possible": 1,
                "due_at": "2023-11-15T23:59:59Z",
                "submission_types": ["none"],
                "published": False,
              }
            }

        Returns:
            requests.Response: The HTTP response from the Canvas API
        """
        return self.session.put(
            url=urljoin(
                settings.CANVAS_BASE_URL,
                f"/api/v1/courses/{self.canvas_course_id}/assignments/{assignment_id}",
            ),
            json=payload,
        )

    def delete_canvas_assignment(self, assignment_id):
        """
        Delete an assignment with the given ID in Canvas.

        Args:
            assignment_id (int): ID of the Canvas Assignment
        """
        return self.session.delete(
            url=urljoin(
                settings.CANVAS_BASE_URL,
                f"/api/v1/courses/{self.canvas_course_id}/assignments/{assignment_id}",
            ),
        )

    def update_assignment_grades(self, canvas_assignment_id, payload):
        return self.session.post(
            url=urljoin(
                settings.CANVAS_BASE_URL,
                f"/api/v1/courses/{self.canvas_course_id}/assignments/{canvas_assignment_id}/submissions/update_grades",
            ),
            data=payload,
        )


def create_assignment_payload(subsection_block):
    """
    Create a Canvas assignment dict matching a subsection block on edX

    Args:
        subsection_block (openedx.core.djangoapps.content.block_structure.block_structure.BlockData):
            The block data for the graded assignment/exam (in the structure of a course, this unit is a subsection)

    Returns:
        dict:
            Assignment payload to be sent to Canvas to create or update the assignment
    """  # noqa: E501
    return {
        "assignment": {
            "name": subsection_block.display_name,
            "integration_id": str(subsection_block.location),
            "grading_type": "percent",
            "points_possible": DEFAULT_ASSIGNMENT_POINTS,
            "due_at": (
                None
                if not subsection_block.fields.get("due")
                # The internal API gives us a TZ-naive datetime for the due date, but Studio indicates that  # noqa: E501
                # the user should enter a UTC datetime for the due date. Coerce this to UTC before creating the  # noqa: E501
                # string representation.
                else subsection_block.fields["due"].astimezone(pytz.UTC).isoformat()
            ),
            "submission_types": ["none"],
            "published": False,
        }
    }


def update_grade_payload_kv(user_id, grade_percent):
    """
    Returns a key/value pair that will be used in the body of a bulk grade update request

    Args:
        user_id (int): The Canvas user ID
        grade_percent (numpy.float64): The percent score of the grade (between 0 and 1)

    Returns:
        (tuple): A key/value pair that will be used in the body of a bulk grade update request
    """  # noqa: D401, E501
    return (f"grade_data[{user_id}][posted_grade]", f"{grade_percent * 100}%")



"""
Canvas Integration API endpoint urls.
"""

from django.urls import re_path

from ol_openedx_canvas_integration import views

urlpatterns = [
    re_path(
        r"^add_canvas_enrollments$",
        views.add_canvas_enrollments,
        name="add_canvas_enrollments",
    ),
    re_path(
        r"^list_canvas_enrollments$",
        views.list_canvas_enrollments,
        name="list_canvas_enrollments",
    ),
    re_path(
        r"^list_canvas_assignments$",
        views.list_canvas_assignments,
        name="list_canvas_assignments",
    ),
    re_path(
        r"^list_canvas_grades$", views.list_canvas_grades, name="list_canvas_grades"
    ),
    re_path(r"^push_edx_grades$", views.push_edx_grades, name="push_edx_grades"),
]



"""
Module containing the Celery tasks that are called from the CMS.

This module ideally should be merged with tasks.py. However, tasks.py
contains references to settings which are not present in the 'cms'.
Since cms dynamically loads handlers.py via ready() in app config,
making handler.py depend on tasks.py results in initialization failures.
So, this module has been added without any dependencies on LMS only
settings.
"""

from __future__ import annotations

import logging

import requests
from celery import shared_task
from lms.djangoapps.courseware.courses import get_course_by_id
from opaque_keys.edx.locator import CourseLocator

from ol_openedx_canvas_integration.api import course_graded_items
from ol_openedx_canvas_integration.client import CanvasClient, create_assignment_payload
from ol_openedx_canvas_integration.utils import get_canvas_course_id

logger = logging.getLogger(__name__)


def diff_assignments(openedx_assignments, canvas_assignments_map):
    """Perform a diff between the assignments in Canvas and Open edX.

    Args:
        openedx_assignments (list): List of Open edX subsection objects
        canvas_assignments_map (dict): Map of assignment integration IDs to Canvas
                                       assignment IDs

    Returns:
        dict: The diff between the assignments with the following structure:
            {
                "add": [dict]: List of assignment payloads to be created in Canvas,
                "update": {int: dict}: Map of Canvas IDs to assignment payloads,
                "delete": [int]: List of Canvas assignment IDs to delete
            }
    """
    assignment_diff = {"add": [], "update": {}, "delete": []}
    for subsection in openedx_assignments:
        integration_id = str(subsection.location)
        payload = create_assignment_payload(subsection)
        canvas_assignment = canvas_assignments_map.pop(integration_id, None)
        if canvas_assignment:
            # if the assignment exists in Canvas, remove from the map to indicate
            # it's synced.
            payload["assignment"]["published"] = canvas_assignment.get(
                "is_published", False
            )
            assignment_diff["update"][canvas_assignment["id"]] = payload
        else:
            assignment_diff["add"].append(payload)

    # any left over assignments in the map is considered as deleted in Open edX
    assignment_diff["delete"] = [c["id"] for c in canvas_assignments_map.values()]

    return assignment_diff


def add_assignments(canvas, assignment_payloads: list[dict]):
    """Add new assignments in Canvas from provided Open edX assignments.

    Args:
        canvas: Canvas client instance used to interact with the Canvas API
        assignment_payloads (list[dict]): List of Canvas assignment payload dict
    """

    succeeded = 0
    for payload in assignment_payloads:
        res = canvas.create_canvas_assignment(payload)
        try:
            res.raise_for_status()
            succeeded += 1
        except requests.HTTPError as e:
            logger.warning(
                "Failed to create new assignment for subsection: %s. Error: %s",
                payload["assignment"]["integration_id"],
                str(e),
            )
    if succeeded:
        logger.info(
            "%d of %d new assignments were successfully added in Canvas.",
            succeeded,
            len(assignment_payloads),
        )


def update_assignments(canvas, assignment_map: dict[int, dict]):
    """Update existing assignments in Canvas with the latest data.

    Args:
        canvas: Canvas client instance used to interact with the Canvas API
        assignment_map (dict[int, dict]): Map of Canvas assignment IDs to payload dicts
    """
    succeeded = 0
    for canvas_id, payload in assignment_map.items():
        res = canvas.update_canvas_assignment(canvas_id, payload)
        try:
            res.raise_for_status()
            succeeded += 1
        except requests.HTTPError as e:
            logger.warning(
                "Failed to update Canvas Assignment %d. Error: %s", canvas_id, str(e)
            )
    if succeeded:
        logger.info(
            "%d of %d assignments were successfully updated in Canvas.",
            succeeded,
            len(assignment_map),
        )


def delete_assignments(canvas, assignment_ids: list[int]):
    """Delete given assignments from Canvas.

    Args:
        canvas: Canvas client instance used to interact with the Canvas API
        assignment_ids (list[int]): List of Canvas assignment IDs to delete
    """
    succeeded = 0
    for canvas_id in assignment_ids:
        res = canvas.delete_canvas_assignment(canvas_id)
        try:
            res.raise_for_status()
            succeeded += 1
        except requests.HTTPError as e:
            logger.warning(
                "Failed to delete Canvas assignment: %d. Error %s", canvas_id, str(e)
            )
    if succeeded:
        logger.info(
            "%d for %d assignments were successfully deleted in Canvas.",
            succeeded,
            len(assignment_ids),
        )


@shared_task
def sync_course_assignments_with_canvas(course_id):
    """
    Sync the assignments in the course with Canvas if there is a linked Canvas course.

    Args:
        course_id (str): Open edX Course ID
    """
    course_key = CourseLocator.from_string(course_id)
    course = get_course_by_id(course_key)
    canvas_course_id = get_canvas_course_id(course)

    if not canvas_course_id:
        logger.info(
            "Course %s is not mapped to a Canvas Course. Skipping assignment sync.",
            course_id,
        )
        return

    openedx_assignments = [
        item["subsection_block"] for _, item, _ in course_graded_items(course)
    ]
    canvas = CanvasClient(canvas_course_id=canvas_course_id)
    canvas_assignments = canvas.get_canvas_assignments()

    operations_map = diff_assignments(openedx_assignments, canvas_assignments)
    logger.info(
        "Syncing assignments with Canvas. Adding: %d, Updating: %d, Deleting: %d",
        len(operations_map["add"]),
        len(operations_map["update"]),
        len(operations_map["delete"]),
    )

    add_assignments(canvas, operations_map["add"])
    update_assignments(canvas, operations_map["update"])
    delete_assignments(canvas, operations_map["delete"])



"""Production settings unique to canvas integration plugin."""

from path import Path as path  # noqa: N813

PLUGIN_TEMPLATES_ROOT = path(__file__).abspath().dirname().dirname()


def plugin_settings(settings):
    """Settings for the canvas integration plugin."""  # noqa: D401
    settings.CANVAS_ACCESS_TOKEN = settings.AUTH_TOKENS.get(
        "CANVAS_ACCESS_TOKEN", settings.CANVAS_ACCESS_TOKEN
    )
    settings.CANVAS_BASE_URL = settings.ENV_TOKENS.get(
        "CANVAS_BASE_URL", settings.CANVAS_BASE_URL
    )

    settings.TEMPLATES = settings.ENV_TOKENS.get("TEMPLATES", settings.TEMPLATES)

    for template_engine in settings.TEMPLATES:
        template_dirs = template_engine["DIRS"]
        template_dirs.append(PLUGIN_TEMPLATES_ROOT + "/templates")
